
// File: <graphics_opengl.h>
// Copyright (c) 1997-2007, Reachin Technologies AB.
// All rights reserved.
// The copyright to the computer program herein is the property of Reachin
// Technologies AB, Sweden. The program may be used and/or copied only with the
// written permission of Reachin Technologies AB or in accordance with the terms
// stipulated in the agreement under which the program has been supplied.

/** Edited by Rolf Anders Syvertsen
	04.05.2007
	rolfans@stud.ntnu.no
*/

#ifndef _GRAPHICS_OPENGL_H_
#define _GRAPHICS_OPENGL_H_

#include "graphics_framework.h"

#include <list>
#include <stdio.h>
#include <stdarg.h>
#include <gl/gl.h>
#include <gl/glu.h>

#pragma comment(lib,"opengl32.lib")
#pragma comment(lib,"glu32.lib")

class TextureOpenGL : public IGraphicsTexture
{
private:
	GLuint m_textureID;
	image_t m_imageCopy;

public:
	TextureOpenGL(image_t& image);
	virtual ~TextureOpenGL();

	void				Bind();
	void				Unbind();
	void				Reload();
	IGraphicsTexture*	Clone(IGraphicsTexture* obj);
	image_t				GetImage();
};


class RenderModelOpenGL : public IGraphicsRenderModel
{
public:
	float m_transform[16];

public:
	RenderModelOpenGL();
	virtual ~RenderModelOpenGL();

	void	Render(bool useMaterials);
	void	Reload();
	void	SetTransformMatrix(const float* mat);
	void	SetTransformMatrix(float pos_x, float pos_y, float pos_z,
							   float rot_x, float rot_y, float rot_z,
							   float scl_x, float scl_y, float scl_z);
	float*	GetTransformMatrix();
	void	PushTransform();
	void	PopTransform();
};


class GraphicsOpenGL : public IGraphicsFramework
{
private:
	int m_width;
	int m_height;
	float m_lightPos[4];
	float m_lightCol[4];
	float m_cameraMatrix[16];
	std::list<IGraphicsRenderModel*> m_modelList;
	HGLRC m_hRC;
	HDC m_hDC;

	GLuint m_fontBase;
	int m_fontSize;

	float m_camRot[2];
	float m_camPos[3];

private:
	void InitFont();
	void DestroyFont();
	void UpdateCamera();

public:
	GraphicsOpenGL();
	virtual ~GraphicsOpenGL();

	const char*				GetAPIName();
	bool					Init(HWND hwnd, int width, int height, bool fullscreen);
	void					Destroy();
	void					Resize(int width, int height);
	void					SetupPerspectiveProjection(float fov, float aspect, float znear, float zfar);

	void					SetLighting(bool isEnabled);
	void					SetCulling(bool isEnabled);
	void					SetDepthMask(bool isEnabled);
	void					SetDepthTest(bool isEnabled);

	IGraphicsTexture*		LoadTexture(const char* strFile);
	IGraphicsRenderModel*	CreateRenderModel();

	bool					IsRightHanded();

	void					AddRenderModel(IGraphicsRenderModel* model);
	void					RemoveRenderModel(IGraphicsRenderModel* model);
	void					RemoveAllRenderModels();
	void					RenderScene(bool useMaterials);
	void					RenderLineArray(float* verts, int size);
	void					RenderBegin();
	void					RenderFinish();

	float*					GetCameraMatrix();
	void					CameraGetPosition(float& x, float& y, float& z);
	void					CameraReset();
	void					CameraMoveForwardBackward(float val, bool add_y);
	void					CameraMoveUpDown(float val);
	void					CameraMoveLeftRight(float val);
	void					CameraRotate(float ang_x, float ang_y);
	void					SetCamera(const float* matrix);
	void					SetCamera(float rix, float riy, float riz,
									  float upx, float upy, float upz,
									  float drx, float dry, float drz,
									  float px, float py, float pz);

	void					SetLightPosition(float x, float y, float z, float w);
	void					SetLightColor(float r, float g, float b, float a);

	void					SetCullBackFace(bool cull_backfaces);

	void					SetColor(float r, float g, float b, float a);

	int						GetFontHeight();
	void					PrintText(int x, int y, const char* str, ...);
};

#endif