
/** This file is based on the HaptX example kode
	Edited by Rolf Anders Syvertsen
	04.05.2007
	rolfans@stud.ntnu.no
	rolfans@gmail.com
*/

#include "NeedleInsertion.h"

#include <iostream>
#include <fstream>
#include <string>
#include <HaptX.h>

#define LOGNAME L"HaptXtest.log"

HaptX::iHapticShapeID	haptx_shapeID = 0;
// only 12 surfaces, that should hold for now
HaptX::SurfaceInfo*		haptx_surfaceInfo[12];
HaptX::HaptXInterface*	haptx_interface = 0;
HaptX::iHapticDeviceID	haptx_deviceID = -1;	// note that 0 is a valid device ID
HaptX::Vectors::Vec3f	haptx_InitPosition(0.0,0.0,0.0);
HaptX::Vectors::Vec3f	haptx_LastPosition(0.0,0.0,0.0);
HaptX::Vectors::Vec3f	haptx_LastAngles(0.0,0.0,0.0);
std::string				strNavigationModel;

int HapticDeviceEnumFunc(std::wstring* deviceName, std::wstring* deviceModel)
{
	if (!haptx_interface || !deviceName  || !deviceModel) return HaptX::HX_STOP;
	haptx_interface->DeviceCreate(haptx_deviceID,deviceName,deviceModel);
	return HaptX::HX_STOP;
}

void HaptXMsgCallback(int Severity, std::wstring Message)
{
	static std::wstring old_message = L"";
	if (Message == old_message) return;
	old_message = Message;

	std::wofstream out(LOGNAME, std::ios::out | std::ios::app);
	if (!out.is_open()) throw std::ios::failure(std::string("Error opening logfile for writing."));
	
		 if (Severity >= HX_ERR_MSG)	out << "Error: ";
	else if (Severity >= HX_WARN_MSG)	out << "Warning: ";
	else if (Severity >= HX_INFO_MSG)	out << "Information: ";
	else if (Severity >= HX_DBG_MSG)	out << "Debug: ";
	else								out << "Haptic log: ";

	out << Message.c_str();
	out.close();
}

NeedleInsertion::NeedleInsertion()
{
	m_model = 0;
	cx = 0.0f;
	cy = 0.0f;
	m_activeNavigationModel = 0;
	strNavigationModel = "Device position, Relative";
	m_camSensitivity = 50.0f;
	m_prevButton1 = false;
	m_prevButton2 = false;
}

NeedleInsertion::~NeedleInsertion()
{
}

void NeedleInsertion::HandleMoveCamera(float deltaTime)
{
	float moveSpeed = deltaTime * 30.0f;

	if (m_input->keyboardKey['W'])
	{
		m_graphics->CameraMoveForwardBackward(moveSpeed,true);
	}

	if (m_input->keyboardKey['S'])
	{
		m_graphics->CameraMoveForwardBackward(-moveSpeed,true);
	}

	if (m_input->keyboardKey['A'])
	{
		m_graphics->CameraMoveLeftRight(moveSpeed);
	}

	if (m_input->keyboardKey['D'])
	{
		m_graphics->CameraMoveLeftRight(-moveSpeed);
	}
	
	if (m_input->keyboardKey[VK_SPACE])
	{
		m_input->keyboardKey[VK_SPACE] = false;
		m_activeNavigationModel++;
		if (m_activeNavigationModel > 2) m_activeNavigationModel = 0;
		
		m_prevButton1 = false;
		m_prevButton2 = false;
		haptx_interface->HaptXEnableShapeRendering();

		if (m_activeNavigationModel == 0)
		{
			strNavigationModel = "Device position, Relative";
			m_camSensitivity = 50.0f;
		}

		if (m_activeNavigationModel == 1)
		{
			strNavigationModel = "Device position, Absolute";
			m_camSensitivity = 10000.0f;
		}

		if (m_activeNavigationModel == 2) 
		{
			strNavigationModel = "Device orientation, Relative";
			m_camSensitivity = 1000.0f;
		}
	}
}

void NeedleInsertion::HandleCameraRotation(float x, float y)
{
	cx -= y * 0.000001f;
	cy -= x * 0.000001f;

	m_graphics->CameraRotate(cx,cy);
}

const char* NeedleInsertion::GetDemoName()
{
	return "HaptX Needle Insertion on Tetrahedral Mesh by Rolf Anders";
}

bool NeedleInsertion::InitDemo(IGraphicsFramework* gfx, inputState_t* inputState)
{
	// load models and setup rendering
	std::wofstream out(LOGNAME, std::ios::out | std::ios::trunc); //Clear any existing log file from any previous run.
	out.close();
	m_graphics = gfx;
	m_input = inputState;
	// the tetmesh
	m_ptetmesh=new TetMesh();
	// the syringe 
//	m_proxyModel = m_graphics->LoadRenderModel("Syringe.ASE");
//	m_syringe = m_graphics->LoadRenderModel("Syringe.ASE");
//	m_syringe->SetTransformMatrix(0.0f,00.0f,0.0f,0.0f,0.0f,0.0f,0.0f,0.0f,0.0f);
//	m_graphics->AddRenderModel(m_syringe);
	
	m_floor = m_graphics->LoadRenderModel("plane_large.ASE");
	m_graphics->AddRenderModel(m_floor);
	m_floor->SetTransformMatrix(0.0f,-40.0f,0.0f,0.0f,0.0f,0.0f,1.5f,1.5f,1.5f);

	m_proxyModel = m_graphics->LoadRenderModel("proxy.ASE");
	m_graphics->AddRenderModel(m_proxyModel);
	m_graphics->SetLightPosition(-40.0f,30.0f,40.0f,1.0f);
	m_graphics->SetLighting(true);
	m_graphics->CameraMoveForwardBackward(-30.0f,false);

	// Loading the tetrahedral mesh model
	LoadingFrame("- Loading the tetrahedral mesh");
	if (!m_ptetmesh->loadTetMesh("shoulder.1")) 
//	if (!m_ptetmesh->loadTetMesh("bar4.1")) 
		throw "Failed to load tetrahedral mesh from .ele file. \n ";
	LoadingFrame("Making the rendrer model from the tetrahedral mesh");
	m_model=m_graphics->LoadRenderModel(m_ptetmesh);

	m_graphics->AddRenderModel(m_model);
	m_model->SetTransformMatrix(0.0f,-160.0f,-90.0f,-1.5f,0.0f,0.0f);


	LoadingFrame("Setting up the haptic scene");
	// initialize the HaptX Interface
	haptx_interface = HaptX::GetHaptXInterface();
	if (!haptx_interface) throw "Failed to retrieve HaptX interface";

	if (haptx_interface->HaptXInit() != HX_SUCCESS)
		throw "HaptXInit failed";
	
	haptx_interface->HaptXHookMessages(HX_INFO_MSG,HaptXMsgCallback); //Register to recieve all messages with a severity down to and including Information messages. 
	haptx_interface->AllDevicesEnum(HapticDeviceEnumFunc);
	if (haptx_deviceID < 0) throw "Failed to create a device. Please make sure the Touch Controller is connected.";

	haptx_interface->DeviceSetWorldScale(haptx_deviceID,0.005f);
	haptx_interface->HaptXSetRightHanded();

	// enable camera oriented haptics
	haptx_interface->DeviceSetCameraRelative(haptx_deviceID,true);

	// create the surfaces for our shapes
	// need to be changed and fixed
	haptx_interface->SurfaceCreateSimple(&(haptx_surfaceInfo[0]));
	haptx_interface->SurfaceCreateSimple(&(haptx_surfaceInfo[1]));
	haptx_interface->SurfaceCreateSimple(&(haptx_surfaceInfo[2]));
	haptx_interface->SurfaceCreateSimple(&(haptx_surfaceInfo[3]));
	haptx_interface->SurfaceCreateSimple(&(haptx_surfaceInfo[4]));
	haptx_interface->SurfaceCreateSimple(&(haptx_surfaceInfo[5]));
	haptx_interface->SurfaceCreateSimple(&(haptx_surfaceInfo[6]));
	haptx_interface->SurfaceCreateSimple(&(haptx_surfaceInfo[7]));
	haptx_interface->SurfaceCreateSimple(&(haptx_surfaceInfo[8]));
	haptx_interface->SurfaceCreateSimple(&(haptx_surfaceInfo[9]));
	haptx_interface->SurfaceCreateSimple(&(haptx_surfaceInfo[10]));


	// create the haptic shapes
	CreateHapticsShape();

	return true;
}
void NeedleInsertion::CreateHapticsShape()
{
	for(int k=0;k<m_model->m_objectArray.size();k++)
	{
		LoadingFrame(" - Creating haptic shape ");
		size_t i;
		size_t haptx_numVertices = (size_t)m_model->m_objectArray[k].m_numVertices;
		size_t haptx_numIndexes = (size_t)m_model->m_objectArray[k].m_numFaces * 3;		// all faces are triangles
		HaptX::Vectors::Vec3f* haptx_vertices = new HaptX::Vectors::Vec3f[haptx_numVertices];
		int* haptx_indexes = new int[haptx_numIndexes];

		for (i = 0; i < haptx_numVertices; i++)
		{
			haptx_vertices[i].x = m_model->m_objectArray[k].m_vertices[i].vertex[0];
			haptx_vertices[i].y = m_model->m_objectArray[k].m_vertices[i].vertex[1];
			haptx_vertices[i].z = m_model->m_objectArray[k].m_vertices[i].vertex[2];
		}

		for (i = 0; i < haptx_numIndexes; i++)
		{
			haptx_indexes[i] = m_model->m_objectArray[k].m_faces[i];
		}

		LoadingFrame(" - Creating haptic shape connecting to surface");
		haptx_interface->ShapeCreate(haptx_shapeID,haptx_vertices,haptx_numVertices,haptx_indexes,haptx_numIndexes,haptx_surfaceInfo[k]->GetSurfaceID());

		// set shape transformation matrix
		float* mt = m_model->GetTransformMatrix();
		haptx_interface->ShapeSetTransform(haptx_shapeID,
										   mt[ 0],mt[ 4],mt[ 8],mt[12],
										   mt[ 1],mt[ 5],mt[ 9],mt[13],
										   mt[ 2],mt[ 6],mt[10],mt[14],
										   mt[ 3],mt[ 7],mt[11],mt[15]);

		
		m_model->m_objectArray[k].m_haptx_shapeID=haptx_shapeID;
		delete [] haptx_vertices;
		delete [] haptx_indexes;
	}
}
void NeedleInsertion::ShutDownDemo()
{
	if (m_proxyModel) delete m_proxyModel;
	if (m_model) delete m_model;
	if (m_floor) delete m_floor;
	if (haptx_interface) haptx_interface->DeviceDestroy(haptx_deviceID);
}

void NeedleInsertion::UpdateDemo(float deltaTime, int width, int height)
{
	// has the device been disconnected?
	bool devConnected = false;
	haptx_interface->DeviceIsConnected(devConnected,haptx_deviceID);
	if (!devConnected) throw "The device has been disconnected";

	// handle camera movement (keys: W, A, S, D)
	HandleMoveCamera(deltaTime);

	// handle camera rotation
	bool button1, button2;
	haptx_interface->DeviceGetButton(button1,haptx_deviceID,0);
	haptx_interface->DeviceGetButton(button2,haptx_deviceID,1);

	bool button1Pressed = (button1 && !m_prevButton1);
	bool button2Pressed = (button2 && !m_prevButton2);
	bool button1Released = (!button1 && m_prevButton1);
	bool button2Released = (!button2 && m_prevButton2);

	HaptX::Vectors::Vec3f devicePosition;
	haptx_interface->DeviceGetHardwarePosition(devicePosition,haptx_deviceID);
	HaptX::Vectors::Vec3f deviceAngles;
	haptx_interface->DeviceGetYawPitchRoll(deviceAngles,haptx_deviceID);

	if (button1 || button2)
	{
		HaptX::Vectors::Vec3f camRotation = HaptX::Vectors::Vec3f(0.0,0.0,0.0);

		if (m_activeNavigationModel == 0 || m_activeNavigationModel == 1)
		{
			if (m_activeNavigationModel == 1)
			{
				haptx_InitPosition  = haptx_LastPosition;
				if (button1Pressed || button2Pressed)
				{
					haptx_interface->HaptXDisableShapeRendering();
				}
			}
			else
			{
				if (button1Pressed || button2Pressed)
				{
					haptx_InitPosition = devicePosition;
				}
			}

			HaptX::Vectors::Vec3f difference = devicePosition  -  haptx_InitPosition;
			
			camRotation.x -= difference.x;
			camRotation.y += difference.y;
			camRotation.z += difference.z;
		}
		else if (m_activeNavigationModel == 2)
		{
			camRotation = deviceAngles;
		}

		HaptX::Vectors::Vec3f camMovement = m_camSensitivity * camRotation;
		HandleCameraRotation((float)camMovement.x ,(float) camMovement.y);

	}
	if (button1Released || button2Released)
	{
		haptx_interface->HaptXEnableShapeRendering();
	}

	m_prevButton1 = button1;
	m_prevButton2 = button2; 
	haptx_LastPosition = devicePosition;
	haptx_LastAngles = deviceAngles;


	// send the camera matrix into the haptic interface
	float* matr = m_graphics->GetCameraMatrix();
	haptx_interface->DeviceSetCameraMatrix(haptx_deviceID,
										   matr[0],matr[4],matr[8],
										   matr[1],matr[5],matr[9],
										   matr[2],matr[6],matr[10],
										   -matr[12],-matr[13],-matr[14],
										   0,0,-30.0f);

	// update the device
	haptx_interface->DeviceRender(haptx_deviceID);

	// retrieve the proxy position
	HaptX::Vectors::Vec3f pos;
	haptx_interface->DeviceGetProxyPosition(pos,haptx_deviceID);
	m_proxyModel->SetTransformMatrix((float)pos.x,(float)pos.y,(float)pos.z,0.0f,0.0f,0.0f,0.05f,0.05f,0.05f);

	// render the scene
	m_graphics->RenderBegin();
//	m_graphics->RenderScene(true); // can't use this with the drawing of the selected polygon
	m_model->Render(true);
	m_floor->Render(true);
	m_graphics->PrintText(2,1,"Input Keys:");
	m_graphics->PrintText(2,1 * (4 + m_graphics->GetFontHeight()),"Esc = Exit Needle Insertion");
	m_graphics->PrintText(2,2 * (4 + m_graphics->GetFontHeight()),"W, A, S, D = Move Camera");
	m_graphics->PrintText(2,3 * (4 + m_graphics->GetFontHeight()),"Space = Change Navigation Mode");
	m_graphics->PrintText(2,4 * (4 + m_graphics->GetFontHeight()),"Press any Touch Controller button to rotate the view");
	m_graphics->PrintText(2,6 * (4 + m_graphics->GetFontHeight()),"Current Navigation Model = %s",strNavigationModel.c_str());

	// detect collision polygons and render them
	// from the polygon picking demo
	bool in_contact = false;
	for(int k=0;k<m_model->m_objectArray.size();k++)
	{
		haptx_interface->ShapeIsDeviceInContact(in_contact,haptx_deviceID,m_model->m_objectArray[k].m_haptx_shapeID);
		if (in_contact)
		{
			// we have a collision, retrieve the face ID's, we can collide with up to three triangles at once
			int ids[3];
			float va[18];

			haptx_interface->ShapeGetContactTriangleIDs(ids[0],ids[1],ids[2],haptx_deviceID,m_model->m_objectArray[k].m_haptx_shapeID);

			// for each of the three triangles that we can collide with
			for (int i = 0; i < 3; i++)
			{
				if (ids[i] != -1)
				{
					// Line 1
					va[ 0] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 0]].vertex[0];
					va[ 1] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 0]].vertex[1];
					va[ 2] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 0]].vertex[2];

					va[ 3] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 1]].vertex[0];
					va[ 4] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 1]].vertex[1];
					va[ 5] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 1]].vertex[2];

					// Line 2
					va[ 6] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 1]].vertex[0];
					va[ 7] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 1]].vertex[1];
					va[ 8] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 1]].vertex[2];
				
					va[ 9] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 2]].vertex[0];
					va[10] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 2]].vertex[1];
					va[11] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 2]].vertex[2];

					// Line 3
					va[12] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 2]].vertex[0];
					va[13] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 2]].vertex[1];
					va[14] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 2]].vertex[2];
				
					va[15] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 0]].vertex[0];
					va[16] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 0]].vertex[1];
					va[17] = m_model->m_objectArray[k].m_vertices[m_model->m_objectArray[k].m_faces[ids[i] * 3 + 0]].vertex[2];

					m_graphics->SetDepthTest(false);
					m_model->PushTransform();
					m_graphics->SetColor(0.0f,1.0f,0.0f,0.0f);
					m_graphics->RenderLineArray(va,6);
					m_graphics->SetColor(1.0f,1.0f,1.0f,0.0f);
					m_model->PopTransform();
					m_graphics->SetDepthTest(true);
				}
				else
				{
					break;
				}
			}
		}
	}
	m_proxyModel->Render(true);
	m_graphics->RenderFinish();
}
void NeedleInsertion::LoadingFrame(std::string message)
{
	// render the scene
	m_graphics->RenderBegin();
	m_floor->Render(true);
//	m_syringe->Render(true);
	m_graphics->PrintText(2,9 * (4 + m_graphics->GetFontHeight()),"Please wait, loading : %s",message.c_str());
	m_graphics->RenderFinish();

}
IDemoFramework* needleIns = new NeedleInsertion();

#include "main.cpp"