
// File: <main.cpp>
// Copyright (c) 1997-2007, Reachin Technologies AB.
// All rights reserved.
// The copyright to the computer program herein is the property of Reachin
// Technologies AB, Sweden. The program may be used and/or copied only with the
// written permission of Reachin Technologies AB or in accordance with the terms
// stipulated in the agreement under which the program has been supplied.

#include <stdexcept>
#include <stdlib.h>
#include <windows.h>
#include "haptX/demo_framework.h"
#include "haptX/graphics_opengl.h"
#include "haptX/resource.h"

#define DEMO_CLASS_NAME		"HaptXTest"

const int				windowWidth = 800;						// window width
const int				windowHeight = 600;						// window height
const bool				windowFullscreen = false;				// fullscreen?

IGraphicsFramework*		graphics = new GraphicsOpenGL();		// "new GraphicsOpenGL()" or "new GraphicsDirectX()"

inputState_t			inputState;
float					oldTime = 0;
float					newTime = 0;
HDC						hdc = 0;
HWND					hWnd = 0;
HINSTANCE				hInstance;
bool					windowCreated = false;
bool					isRunning = true;
int						newWidth = 0;
int						newHeight = 0;
LARGE_INTEGER			initTime;
LARGE_INTEGER			ticksPerSecond;
LARGE_INTEGER			currentTime;


LRESULT	CALLBACK	WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void				ResizeWindow(int width, int height);
void				ShutDown();


bool InitWindow(int width, int height, bool fullscreen)
{
	WNDCLASS wc;
	DWORD dwExStyle;
	DWORD dwStyle;
	RECT WindowRect;
	int px = 0;
	int py = 0;

	// create window
	hInstance = GetModuleHandle(NULL);
	wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc = (WNDPROC)WndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon(hInstance,MAKEINTRESOURCE(IDI_ICON1));
	//wc.hIcon = (HICON)LoadImage(GetModuleHandle(NULL),MAKEINTRESOURCE(IDI_ICON1),IMAGE_ICON,16,16,0);
	wc.hCursor = LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground = NULL;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = DEMO_CLASS_NAME;

	if (!RegisterClass(&wc))
		return false;

	if (fullscreen)
	{
		DEVMODE dmScreenSettings;
		memset(&dmScreenSettings,0,sizeof(dmScreenSettings));
		dmScreenSettings.dmSize = sizeof(dmScreenSettings);
		dmScreenSettings.dmPelsWidth = width;
		dmScreenSettings.dmPelsHeight = height;
		dmScreenSettings.dmBitsPerPel = 32;
		dmScreenSettings.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;

		if (ChangeDisplaySettings(&dmScreenSettings,CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
		{
			MessageBox(NULL,"Fullscreen not supported.","Warning",MB_OK | MB_ICONWARNING);
			fullscreen = false;
		}
	}

	if (fullscreen)
	{
		dwExStyle = WS_EX_APPWINDOW;
		dwStyle = WS_POPUP;
	}
	else
	{
		dwExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
		dwStyle = WS_POPUP | WS_BORDER | WS_SYSMENU | WS_CAPTION; // WS_OVERLAPPEDWINDOW;

		// set window position to middle of screen
		if (width > (int)GetSystemMetrics(SM_CXSCREEN)) width = (int)GetSystemMetrics(SM_CXSCREEN);
		if (height > (int)GetSystemMetrics(SM_CYSCREEN)) height = (int)GetSystemMetrics(SM_CYSCREEN);

		px = ((int)GetSystemMetrics(SM_CXSCREEN) - width) / 2;
		py = ((int)GetSystemMetrics(SM_CYSCREEN) - height) / 2;
	}

	windowCreated = true;
	WindowRect.left = 0;
	WindowRect.right = width;
	WindowRect.top = 0;
	WindowRect.bottom = height;
	AdjustWindowRectEx(&WindowRect,dwStyle,FALSE,dwExStyle);

	if (!(hWnd = CreateWindowEx(dwExStyle,DEMO_CLASS_NAME,"HaptX Needle Insertion by Rolf Anders",dwStyle | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,px,py,width,height,NULL,NULL,hInstance,NULL)))
		return false;

	hdc = GetDC(hWnd);
	if (!hdc) return false;

	ShowWindow(hWnd,SW_SHOW);
	SetForegroundWindow(hWnd);
	SetFocus(hWnd);
	SetCursorPos(px + (width / 2),py + (height / 2));

	if (graphics) 
	{
		if (!graphics->Init(hWnd,width,height,fullscreen))
			return false;

		graphics->SetupPerspectiveProjection(90.0f,(float)width / (float)height,1.0f,10000.0f);
	}

	return true;
}

void ResizeWindow(int width, int height)
{
	if (graphics)
	{
		if (height == 0) height = 1;
		graphics->Resize(width,height);
		graphics->SetupPerspectiveProjection(90.0f,(float)width / (float)height,1.0f,10000.0f);

		newWidth = width;
		newHeight = height;
	}
}

void ShutDown()
{
	if (needleIns)
	{
		needleIns->ShutDownDemo();
		delete needleIns;
	}

	if (graphics)
	{
		graphics->Destroy();
		delete graphics;
	}

	// destroy window
	if (!windowCreated) return;
	if (windowFullscreen) ChangeDisplaySettings(NULL,0);
	ReleaseDC(hWnd,hdc);
	DestroyWindow(hWnd);
	UnregisterClass(DEMO_CLASS_NAME,hInstance);
	windowCreated = false;
}

LRESULT	CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
		case WM_SYSCOMMAND:
		{
			switch (wParam)
			{
				case SC_SCREENSAVE:
				case SC_MONITORPOWER:
				return 0;
			}
			break;
		}

		case WM_CLOSE:
		{
			isRunning = false;
			PostQuitMessage(0);
			return 0;
		}

		case WM_KEYDOWN:
		{
			inputState.keyboardKey[wParam] = true;
			return 0;
		}

		case WM_KEYUP:
		{
			inputState.keyboardKey[wParam] = false;
			return 0;
		}

		case WM_LBUTTONDOWN:
		{
			inputState.mouseButton[0] = true;
			break;
		}

		case WM_LBUTTONUP:
		{
			inputState.mouseButton[0] = false;
			break;
		}

		case WM_RBUTTONDOWN:
		{
			inputState.mouseButton[1] = true;
			break;
		}

		case WM_RBUTTONUP:
		{
			inputState.mouseButton[1] = false;
			break;
		}

		case WM_MOUSEMOVE:
		{
			inputState.mouseX = (int)LOWORD(lParam);
			inputState.mouseY = (int)HIWORD(lParam);
			break;
		}

		case WM_SIZE:
		{
			ResizeWindow(LOWORD(lParam),HIWORD(lParam));
			return 0;
		}
	}

	return DefWindowProc(hWnd,uMsg,wParam,lParam);
}

float GetTotalTime()
{
	QueryPerformanceCounter(&currentTime);
	float ret = (float)((double)(currentTime.QuadPart - initTime.QuadPart) / (double)ticksPerSecond.QuadPart);
	return (ret < 0.0f)? -ret : ret;
}

void OutputFPS(float tm)
{
	static int frame = 0;
	static float oldfpstime = tm;

	frame++;
	if (tm - oldfpstime >= 1.0f)
	{
		char wndText[128];
		sprintf_s(wndText,"%s (API = %s, FPS = %d)",(needleIns)? needleIns->GetDemoName():"HaptX Test by Rolf Anders",(graphics)? graphics->GetAPIName():"None",frame);
		SetWindowText(hWnd,wndText);
		frame = 1;
		oldfpstime = tm;
	}
}

int WINAPI WinMain(HINSTANCE hInstance,	HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG msg;
	int res = 1;
	try
	{
		if (!InitWindow(windowWidth,windowHeight,windowFullscreen))
		{
			MessageBox(NULL,"Failed to create window.","Error",MB_OK | MB_ICONERROR);
			ShutDown();
			return 0;
		}

		if (needleIns)
		{
			if (!needleIns->InitDemo(graphics,&inputState))
			{
				MessageBox(NULL,"Failed to initialize needleIns.","Error",MB_OK | MB_ICONERROR);
				ShutDown();
				return 0;
			}

			SetWindowText(hWnd,needleIns->GetDemoName());
		}

		QueryPerformanceFrequency(&ticksPerSecond);
		QueryPerformanceCounter(&initTime);
		oldTime = newTime = GetTotalTime();

		while (isRunning)
		{
			if (PeekMessage(&msg,NULL,0,0,PM_REMOVE))
			{
				if (msg.message == WM_QUIT)
				{
					isRunning = false;
				}
				else
				{
					TranslateMessage(&msg);
					DispatchMessage(&msg);
				}
			}
			else
			{
				if (inputState.keyboardKey[VK_ESCAPE])
				{
					isRunning = false;
				}
				else
				{
					newTime = GetTotalTime();
					OutputFPS(newTime);
					
					if (needleIns)
					{
						needleIns->UpdateDemo(newTime - oldTime,newWidth,newHeight);
					}
					else if (graphics)
					{
						graphics->RenderBegin();
						graphics->RenderScene(true);
						graphics->SetColor(1.0f,1.0f,1.0f,0.0f);
						graphics->PrintText(20,50,"No Demo Is Running!");
						graphics->RenderFinish();
					}
					
					oldTime = newTime;
				}
			}
		}
		res = (int)msg.wParam;
		ShutDown();
	}
	catch (std::exception& e)
	{
		MessageBox(NULL,e.what(),"Exception",MB_OK | MB_ICONERROR);
	}
	catch (const char* emsg)
	{
		MessageBox(NULL,emsg,"Exception",MB_OK | MB_ICONERROR);
	}

	return res;
}