/** made by Rolf Anders Syvertsen
*	2007-04-13
*	rolfans@gmail.com
*/

#include "tetmesh.h"

TetMesh::TetMesh()
{
	m_numTet=0;
	m_ptetList=0;
	m_numVertexes=0;
	m_pvertexList=0;
	m_numTetAttrib=0;
	m_pvertexTet=0;
	m_firstnum=0; 
}
TetMesh::~TetMesh()
{
	delete [] m_ptetList;
	delete [] m_pvertexList;
	delete [] m_pvertexTet;
}
bool TetMesh::loadTetMesh(char *filename)
{
  FILE *infile;
  char innodefilename[FILENAMESIZE];
  char inelefilename[FILENAMESIZE];
  char inputline[INPUTLINESIZE];
  char *stringptr, *infilename;
  int attrib;
  int markers, numberofpointattributes,numberofcorners;
  int corner;
  int i, j;
  // the tissue types
  vector<int>tissuetypes;


  // Assembling the actual file names we want to open.
  strcpy(innodefilename, filename);
  strcpy(inelefilename, filename);
  strcat(innodefilename, ".node");
  strcat(inelefilename, ".ele");

  // Read the points from a .node file.
  infilename = innodefilename;
  cout<<"Opening "<<infilename<<endl;
  infile = fopen(infilename, "r");
  if (infile == (FILE *) NULL) 
  {
    throw "File I/O Error:  Cannot access file %c",infilename;
    return false;
  }
  // Read number of points, number of dimensions, number of point
  //   attributes, and number of boundary markers.
  stringptr = readnumberline(inputline, infile, infilename);
  m_numVertexes = (int) strtol (stringptr, &stringptr, 0);
  stringptr = findnextnumber(stringptr);
  if (*stringptr == '\0') {
    m_meshdim = 3;
  } else {
    m_meshdim = (int) strtol (stringptr, &stringptr, 0);
  }
  stringptr = findnextnumber(stringptr);
  if (*stringptr == '\0') {
    numberofpointattributes = 0;
  } else {
    numberofpointattributes = (int) strtol (stringptr, &stringptr, 0);
  }
  stringptr = findnextnumber(stringptr);
  if (*stringptr == '\0') {
    markers = 0;  // Default value.
  } else {
    markers = (int) strtol (stringptr, &stringptr, 0);
  }

  if (m_meshdim != 3) {
    throw "Error:  Can only load tetrahedral meshes with 3D points.";
    fclose(infile);
    return false;
  }
  if (m_numVertexes < 4) {
    throw "File I/O error:  Input should have at least 4 points.";
    fclose(infile);
    return false;
  }

  // Load the list of nodes.
  if (!loadNodes(infile, markers, infilename)) {
    fclose(infile);
    return false;
  }
  fclose(infile);

  // Read the elements from an .ele file.
  infilename = inelefilename;
  infile = fopen(infilename, "r");
  if (infile != (FILE *) NULL) 
  {
    // Read number of elements, number of corners (4 or 10), number of
    //   element attributes.
    stringptr = readnumberline(inputline, infile, infilename);
    m_numTet = (int) strtol (stringptr, &stringptr, 0);
    stringptr = findnextnumber(stringptr);
    if (*stringptr == '\0') {
      numberofcorners = 4;  // Default read 4 nodes per element.
    } else {
      numberofcorners = (int) strtol (stringptr, &stringptr, 0);
    }
    stringptr = findnextnumber(stringptr);
    if (*stringptr == '\0') {
      m_numTetAttrib = 0; // Default no attribute.
    } else {
      m_numTetAttrib = (int) strtol (stringptr, &stringptr, 0);
	  if(m_numTetAttrib>1)
	  {
//        cout<<"Error:  Wrong number of tet attributes "<< m_numTetAttrib<<"(should be 0 or 1)."<<endl;
        throw "Error:  Wrong number of tet attributes (should be 0 or 1).";
	  }
    }
	if (numberofcorners != 4 ) //&& numberofcorners != 10) 
    {
//      cout<<"Error:  Wrong number of corners "<< numberofcorners<<"(should be 4)."<<endl;
      throw "Error:  Wrong number of corners (should be 4).";
      fclose(infile);
      return false;
    }
    // Allocate memory for tetrahedra.
    if (m_numTet > 0) {
      m_ptetList = new Tetrahedron[m_numTet]; 
      if (m_ptetList == (Tetrahedron *) NULL) {
         throw "Error:  Out of memory.";
		exit(1);
      }
	}
    // Read the list of tetrahedra.
	// the corner vertex indexes
	int *vert=new int[4];
    for (i = 0; i < m_numTet; i++) 
	{
      // Read tetrahedron index and the tetrahedron's corners.
      stringptr = readnumberline(inputline, infile, infilename);
      for (j = 0; j < numberofcorners; j++) 
	  {
        stringptr = findnextnumber(stringptr);
        if (*stringptr == '\0') 
		{
           cout<<"Error:  Tetrahedron "<< i + m_firstnum<<" is missing vertex "<< j + 1<<" in "<< infilename<<endl;
          exit(1);
        }
        corner = (int) strtol(stringptr, &stringptr, 0);
        if (corner < m_firstnum || corner >= m_numVertexes + m_firstnum) 
		{
          cout<<"Error:  Tetrahedron "<< i + m_firstnum<<" has an invalid vertex index."<<endl;
          exit(1);
        }
		vert[j]=corner;
      }
	  // add the tetrahedrons corners
      m_ptetList[i].addCorners(vert[0]-m_firstnum,vert[1]-m_firstnum,vert[2]-m_firstnum,vert[3]-m_firstnum);
	  // calculate the face normal
	  calcFaceNorm(i);

      // Read the tetrahedron's attributes.
      for (j = 0; j < m_numTetAttrib; j++) 
	  {
        stringptr = findnextnumber(stringptr);
        if (*stringptr == '\0') 
		{
          attrib = 0;
        } else 
		{
		  attrib = (int) strtol(stringptr, &stringptr, 0);
        }
		m_ptetList[i].setType(attrib);
		// counting the tissue types
		  bool tissue=false;
		  for(int k=0;k<tissuetypes.size();k++)
		  {
			  if(tissuetypes.at(k)==attrib)
			  {
				  tissue=true;
				  break;
			  }
		  }
		  if(!tissue)
		  {
			  tissuetypes.push_back(attrib);
		  }
      }
    }
    fclose(infile);
	m_numTissueTypes=tissuetypes.size();
  }
  // connect tets
  connectTet();
  // calc vertex normals
  calculateVertexNorm();
  // find the number of faces in the differen tissue types
  makeDispObjects();

  return true;
}
bool TetMesh::loadNodes(FILE* infile, int markers, char* infilename)
{
  char inputline[INPUTLINESIZE];
  char *stringptr;
  float x, y, z; 
  int firstnode;
  int i;

  // Initialize Vertex List, 'pointattributelist', and 'pointmarkerlist'.
  m_pvertexList = new Vertex[m_numVertexes];
  if (m_pvertexList == (Vertex *) NULL) {
    throw "Error:  Out of memory.";
    exit(1);
  }
  // Read the point section.
  for (i = 0; i < m_numVertexes; i++) {
    stringptr = readnumberline(inputline, infile, infilename);
    if (i == 0) {
      firstnode = (int) strtol (stringptr, &stringptr, 0);
      if ((firstnode == 0) || (firstnode == 1)) {
        m_firstnum = firstnode;
      }
    }
    stringptr = findnextnumber(stringptr);
    if (*stringptr == '\0') {
      cout<<"Error:  Point "<<m_firstnum + i <<" has no x coordinate."<<endl;
      break;
    }
    x = (float) strtod(stringptr, &stringptr);
    stringptr = findnextnumber(stringptr);
    if (*stringptr == '\0') {
       cout<<"Error:  Point "<< m_firstnum + i<<" has no y coordinate."<<endl;
      break;
    }
    y = (float) strtod(stringptr, &stringptr);
    if (m_meshdim == 3) {
      stringptr = findnextnumber(stringptr);
      if (*stringptr == '\0') {
        cout<<"Error:  Point "<< m_firstnum + i<<" has no z coordinate."<<endl;
        break;
      }
      z = (float) strtod(stringptr, &stringptr);
    } else {
      z = 0.0; // m_meshdim == 2;
    }
	m_pvertexList[i].setPoint(x,y,z);

  }
  if (i < m_numVertexes) {
    // Failed to read points due to some error.
    delete [] m_pvertexList;
    m_pvertexList = (Vertex *) NULL;
	throw " Error reading points from .node file.";
    m_numVertexes = 0;
    return false;
  }
  return true;
}
void TetMesh::calcFaceNorm(int tetind)
{	
	// the face normal
	//n = (u x v)/(|u x v|)
	// u = 1 - 0
	// v = 2 - 0
	int *vertind;
	Vertex u; 
	Vertex v; Vertex n1; Vertex n; 
	vertind=m_ptetList[tetind].getVertex();
	// face 0
	u.setPoint(m_pvertexList[vertind[2]].getX()-m_pvertexList[vertind[0]].getX(),
			m_pvertexList[vertind[2]].getY()-m_pvertexList[vertind[0]].getY(),
			m_pvertexList[vertind[2]].getZ()-m_pvertexList[vertind[0]].getZ());
	v.setPoint(m_pvertexList[vertind[1]].getX()-m_pvertexList[vertind[0]].getX(),
			m_pvertexList[vertind[1]].getY()-m_pvertexList[vertind[0]].getY(),
			m_pvertexList[vertind[1]].getZ()-m_pvertexList[vertind[0]].getZ());
	// the cross product
	n1=u.cross(v);
	n=n1*(float)((float)1.0/((float)n1.length()));
	// set the face normal
	m_ptetList[tetind].setFaceNorm(0,n);
	// face 1
	u.setPoint(m_pvertexList[vertind[3]].getX()-m_pvertexList[vertind[0]].getX(),
			m_pvertexList[vertind[3]].getY()-m_pvertexList[vertind[0]].getY(),
			m_pvertexList[vertind[3]].getZ()-m_pvertexList[vertind[0]].getZ());
	v.setPoint(m_pvertexList[vertind[2]].getX()-m_pvertexList[vertind[0]].getX(),
			m_pvertexList[vertind[2]].getY()-m_pvertexList[vertind[0]].getY(),
			m_pvertexList[vertind[2]].getZ()-m_pvertexList[vertind[0]].getZ());
	// the cross product
	n1=u.cross(v);
	n=n1*(float)((float)1.0/((float)n1.length()));
	// set the face normal
	m_ptetList[tetind].setFaceNorm(1,n);
	// face 2
	u.setPoint(m_pvertexList[vertind[1]].getX()-m_pvertexList[vertind[0]].getX(),
			m_pvertexList[vertind[1]].getY()-m_pvertexList[vertind[0]].getY(),
			m_pvertexList[vertind[1]].getZ()-m_pvertexList[vertind[0]].getZ());
	v.setPoint(m_pvertexList[vertind[3]].getX()-m_pvertexList[vertind[0]].getX(),
			m_pvertexList[vertind[3]].getY()-m_pvertexList[vertind[0]].getY(),
			m_pvertexList[vertind[3]].getZ()-m_pvertexList[vertind[0]].getZ());
	// the cross product
	n1=u.cross(v);
	n=n1*(float)((float)1.0/((float)n1.length()));
	// set the face normal
	m_ptetList[tetind].setFaceNorm(2,n);
	// face 3
	u.setPoint(m_pvertexList[vertind[2]].getX()-m_pvertexList[vertind[1]].getX(),
			m_pvertexList[vertind[2]].getY()-m_pvertexList[vertind[1]].getY(),
			m_pvertexList[vertind[2]].getZ()-m_pvertexList[vertind[1]].getZ());
	v.setPoint(m_pvertexList[vertind[3]].getX()-m_pvertexList[vertind[1]].getX(),
			m_pvertexList[vertind[3]].getY()-m_pvertexList[vertind[1]].getY(),
			m_pvertexList[vertind[3]].getZ()-m_pvertexList[vertind[1]].getZ());
	// the cross product
	n1=u.cross(v);
	n=n1*(float)((float)1.0/((float)n1.length()));
	// set the face normal
	m_ptetList[tetind].setFaceNorm(3,n);

}
void TetMesh::connectTet()
{
	if(m_pvertexTet!=0)
		delete [] m_pvertexTet;
	m_pvertexTet=new Vertet[m_numVertexes];
	int *vertid;
	// connecting the tet to the vertexes
	for(int i=0;i<m_numTet;i++)
	{
		vertid=m_ptetList[i].getVertex();
		for(int j=0;j<4;j++)
		{
			m_pvertexTet[vertid[j]].m_tetid.push_back(i);
		}
	}
//	cout<<" connected the vertexes to the tets "<<endl;
	// connecting the tets from the list
	int size=0;
	int *vert, *tempvert;
	int equalVert=0;
	bool *eqvert=new bool[4];
	bool *eqtempvert=new bool[4];
	for(int i=0;i<m_numVertexes;i++)
	{
		size=m_pvertexTet[i].m_tetid.size();
		int tetid1,tetid2;
		for(int j=0;j<size;j++)
		{
			tetid1=m_pvertexTet[i].m_tetid.at(j);
			vert=m_ptetList[tetid1].getVertex();
			for(int k=j+1;k<size;k++)
			{
				tetid2=m_pvertexTet[i].m_tetid.at(k);
				tempvert=m_ptetList[tetid2].getVertex();
				equalVert=0;
				eqvert[0]=false;
				eqvert[1]=false;
				eqvert[2]=false;
				eqvert[3]=false;
				eqtempvert[0]=false;
				eqtempvert[1]=false;
				eqtempvert[2]=false;
				eqtempvert[3]=false;
				for(int l=0;l<4;l++)
				{
					if(vert[0]==tempvert[l])
					{
						equalVert++;
						eqvert[0]=true;
						eqtempvert[l]=true;

					}
					else if(vert[1]==tempvert[l])
					{
						equalVert++;
						eqvert[1]=true;
						eqtempvert[l]=true;
					}
					else if(vert[2]==tempvert[l])
					{
						equalVert++;
						eqvert[2]=true;
						eqtempvert[l]=true;
					}
					else if(vert[3]==tempvert[l])
					{
						equalVert++;
						eqvert[3]=true;
						eqtempvert[l]=true;
					}
				}
				if(equalVert==3)
				{
					// connecting the tetrahedrons
					if(eqvert[0]&&eqvert[1]&&eqvert[2])
					{
						m_ptetList[tetid1].setNeighbor(0,&m_ptetList[tetid2]);
					}
					else if(eqvert[0]&&eqvert[2]&&eqvert[3])
					{
						m_ptetList[tetid1].setNeighbor(1,&m_ptetList[tetid2]);
					}
					else if(eqvert[0]&&eqvert[3]&&eqvert[1])
					{
						m_ptetList[tetid1].setNeighbor(2,&m_ptetList[tetid2]);
					}
					else if(eqvert[1]&&eqvert[3]&&eqvert[2])
					{
						m_ptetList[tetid1].setNeighbor(3,&m_ptetList[tetid2]);
					}
					// connecting back
					if(eqtempvert[0]&&eqtempvert[1]&&eqtempvert[2])
					{
						m_ptetList[tetid2].setNeighbor(0,&m_ptetList[tetid1]);
					}
					else if(eqtempvert[0]&&eqtempvert[2]&&eqtempvert[3])
					{
						m_ptetList[tetid2].setNeighbor(1,&m_ptetList[tetid1]);
					}
					else if(eqtempvert[0]&&eqtempvert[3]&&eqtempvert[1])
					{
						m_ptetList[tetid2].setNeighbor(2,&m_ptetList[tetid1]);
					}
					else if(eqtempvert[1]&&eqtempvert[3]&&eqtempvert[2])
					{
						m_ptetList[tetid2].setNeighbor(3,&m_ptetList[tetid1]);
					}
				}

			}
		}
	}
}
void TetMesh::calculateVertexNorm()
{
	// calculate the vertex normals
	// the vertet has all the sides the vertex is connected to
	// need to check them against the faces that is going to be
	// rendered only !!!!!!!!
	// this is a litle to simple, does not take into account that
	// the vertexes has backsides
	float n[3], l;
	int tetid1;
	int *vertex;
	int size=0;
	for(int i=0;i<m_numVertexes;i++)
	{
		size=m_pvertexTet[i].m_tetid.size();
		n[0]=0.0;
		n[1]=0.0;
		n[2]=0.0;
		for(int j=0;j<size;j++)
		{
			// need to find only the sides with the point that is going
			// to be rendered
			// and then calculating the vertex normal
			tetid1=m_pvertexTet[i].m_tetid.at(j);
			vertex=m_ptetList[tetid1].getVertex();
			if(vertex[0]==i)
			{
				if(m_ptetList[tetid1].renderFaceAt(0))
				{
					n[0]+=m_ptetList[tetid1].getFaceNormAt(0).getX();
					n[1]+=m_ptetList[tetid1].getFaceNormAt(0).getY();
					n[2]+=m_ptetList[tetid1].getFaceNormAt(0).getZ();
				}
				if(m_ptetList[tetid1].renderFaceAt(1))
				{
					n[0]+=m_ptetList[tetid1].getFaceNormAt(1).getX();
					n[1]+=m_ptetList[tetid1].getFaceNormAt(1).getY();
					n[2]+=m_ptetList[tetid1].getFaceNormAt(1).getZ();
				}
				if(m_ptetList[tetid1].renderFaceAt(2))
				{
					n[0]+=m_ptetList[tetid1].getFaceNormAt(2).getX();
					n[1]+=m_ptetList[tetid1].getFaceNormAt(2).getY();
					n[2]+=m_ptetList[tetid1].getFaceNormAt(2).getZ();
				}
			}
			if(vertex[1]==i)
			{
				if(m_ptetList[tetid1].renderFaceAt(0))
				{
					n[0]+=m_ptetList[tetid1].getFaceNormAt(0).getX();
					n[1]+=m_ptetList[tetid1].getFaceNormAt(0).getY();
					n[2]+=m_ptetList[tetid1].getFaceNormAt(0).getZ();
				}
				if(m_ptetList[tetid1].renderFaceAt(2))
				{
					n[0]+=m_ptetList[tetid1].getFaceNormAt(2).getX();
					n[1]+=m_ptetList[tetid1].getFaceNormAt(2).getY();
					n[2]+=m_ptetList[tetid1].getFaceNormAt(2).getZ();
				}
				if(m_ptetList[tetid1].renderFaceAt(3))
				{
					n[0]+=m_ptetList[tetid1].getFaceNormAt(3).getX();
					n[1]+=m_ptetList[tetid1].getFaceNormAt(3).getY();
					n[2]+=m_ptetList[tetid1].getFaceNormAt(3).getZ();
				}
			}
			if(vertex[2]==i)
			{
				if(m_ptetList[tetid1].renderFaceAt(0))
				{
					n[0]+=m_ptetList[tetid1].getFaceNormAt(0).getX();
					n[1]+=m_ptetList[tetid1].getFaceNormAt(0).getY();
					n[2]+=m_ptetList[tetid1].getFaceNormAt(0).getZ();
				}
				if(m_ptetList[tetid1].renderFaceAt(1))
				{
					n[0]+=m_ptetList[tetid1].getFaceNormAt(1).getX();
					n[1]+=m_ptetList[tetid1].getFaceNormAt(1).getY();
					n[2]+=m_ptetList[tetid1].getFaceNormAt(1).getZ();
				}
				if(m_ptetList[tetid1].renderFaceAt(3))
				{
					n[0]+=m_ptetList[tetid1].getFaceNormAt(3).getX();
					n[1]+=m_ptetList[tetid1].getFaceNormAt(3).getY();
					n[2]+=m_ptetList[tetid1].getFaceNormAt(3).getZ();
				}
			}
			if(vertex[3]==i)
			{
				if(m_ptetList[tetid1].renderFaceAt(1))
				{
					n[0]+=m_ptetList[tetid1].getFaceNormAt(1).getX();
					n[1]+=m_ptetList[tetid1].getFaceNormAt(1).getY();
					n[2]+=m_ptetList[tetid1].getFaceNormAt(1).getZ();
				}
				if(m_ptetList[tetid1].renderFaceAt(2))
				{
					n[0]+=m_ptetList[tetid1].getFaceNormAt(2).getX();
					n[1]+=m_ptetList[tetid1].getFaceNormAt(2).getY();
					n[2]+=m_ptetList[tetid1].getFaceNormAt(2).getZ();
				}
				if(m_ptetList[tetid1].renderFaceAt(3))
				{
					n[0]+=m_ptetList[tetid1].getFaceNormAt(3).getX();
					n[1]+=m_ptetList[tetid1].getFaceNormAt(3).getY();
					n[2]+=m_ptetList[tetid1].getFaceNormAt(3).getZ();
				}
			}
		}
		// adding the normal
		l = sqrtf(n[0] * n[0] + n[1] * n[1] + n[2] * n[2]);
		if (l == 0.0f) l = 0.0001f;
		n[0] /= l; n[1] /= l; n[2] /= l;
		m_pvertexList[i].setNormal(n[0],n[1],n[2]);
	}
}
int TetMesh::numTissueTypes()
{
	return m_numTissueTypes;
}
void TetMesh::makeDispObjects()
{
//	if(m_ptissue!=NULL)
//		delete[] m_ptissue;
	m_ptissue=new TissueType[m_numTissueTypes];
	for(int i=0;i<m_numTet;i++)
	{
		int id=0;
		int type=m_ptetList[i].getType();
		for(int j=0;j<m_numTissueTypes;j++)
		{
			int faces=0;
			if(m_ptetList[i].renderFaceAt(0))
				faces++;
			if(m_ptetList[i].renderFaceAt(1))
				faces++;
			if(m_ptetList[i].renderFaceAt(2))
				faces++;
			if(m_ptetList[i].renderFaceAt(3))
				faces++;
			if(faces<=0)
				break;
			if(m_ptissue[j].m_type==type)
			{
				m_ptissue[j].m_numFaces+=faces;
				m_ptissue[j].m_numVertexes+=faces*3;
				break;
			}
			else if(m_ptissue[j].m_type==-9999)
			{
				m_ptissue[j].m_type=type;
				m_ptissue[j].m_numFaces=0;
				m_ptissue[j].m_numVertexes=0;
				m_ptissue[j].m_numFaces+=faces;
				m_ptissue[j].m_numVertexes+=faces*3;
				break;
			}
		}
	}
}
Vertex TetMesh::getVertexAt(int index)
{
	if(index>m_numVertexes||index<0)
		throw "Vertex index is wrong ! /nCan't return this vertex";
	return m_pvertexList[index];
}

char* TetMesh::findnextnumber(char *string)
{
  char *result;

  result = string;
  // Skip the current field.  Stop upon reaching whitespace or a comma.
  while ((*result != '\0') && (*result != '#') && (*result != ' ') && 
         (*result != '\t') && (*result != ',')) {
    result++;
  }
  // Now skip the whitespace and anything else that doesn't look like a
  //   number, a comment, or the end of a line. 
  while ((*result != '\0') && (*result != '#')
         && (*result != '.') && (*result != '+') && (*result != '-')
         && ((*result < '0') || (*result > '9'))) {
    result++;
  }
  // Check for a comment (prefixed with `#').
  if (*result == '#') {
    *result = '\0';
  }
  return result;
}
char* TetMesh::readnumberline(char *string, FILE *infile, char *infilename)
{
  char *result;

  // Search for something that looks like a number.
  do {
    result = fgets(string, INPUTLINESIZE, infile);
    if (result == (char *) NULL) {
      if (infilename != (char *) NULL) {
        cout<<"  Error:  Unexpected end of file in "<<infilename<<endl;
        exit(1);
      }
      return result;
    }
    // Skip anything that doesn't look like a number, a comment, 
    //   or the end of a line. 
    while ((*result != '\0') && (*result != '#')
           && (*result != '.') && (*result != '+') && (*result != '-')
           && ((*result < '0') || (*result > '9'))) {
      result++;
    }
    // If it's a comment or end of line, read another line and try again.
  } 
  while ((*result == '#') || (*result == '\0'));
  return result;
}
