/**	tetrahedral mesh for the HaptX scene graph
	made by Rolf Anders Syvertsen
	2007-04-13
	rolfans@stud.ntnu.no
	rolfans@gmail.com
*/
#ifndef TETMESH_H
#define TETMESH_H

#include <stdio.h>
#include <iostream>
#include <windows.h>
#include<gl/gl.h>
#include<gl/glu.h>

#include <math.h>
#include <vector>

#include "tetrahedron.h"
#include "vertex.h"
#include "vertet.h"
#include "tissuetype.h"

using namespace std; 

/** The loading of the tetrahedral mesh is based on the loader in
*	TetGen. 
*	Loads meshes from .ele and .node files
*/
class TetMesh//:public Geometry
{
public:
	//! Constructor
	TetMesh();
	//! Destructor
	~TetMesh();

	//! load tetrahedral mesh from a .ele and a .node file
	bool loadTetMesh(char *filename);
	//! load the nodes from the .node file
	bool loadNodes(FILE* infile, int markers, char* infilename);
	
	//! calculate face normals for tetrahedron at
	void calcFaceNorm(int tetind);

	//! the number of tissue types
	int numTissueTypes();
	//! the tissueTypes
	TissueType *getTissueData(){return m_ptissue;};

	//! the number of tets
	int numTets(){return m_numTet;}
	//! the tet at
	Tetrahedron *getTetAt(int index){return &m_ptetList[index];}
	//! the vertex at
	Vertex getVertexAt(int index);

	//! types of tissue
	enum TISSUETYPE{SKIN=100, SOFTTISSUE=101, BLOODVEINS=102,
		BLOODARTERIES=103, NERVES=104, LUNG=105, HART=106,
		MEMBRANE=107,
		BONE=200, // should be from 200 -> 299
		MUSCLES=300 // should be from 300 -> 399

	};

protected:
	//! connect the tetrahedrons
	void connectTet();

	//! calculate the vertex normals of the
	//! vertexes that will be rendered
	void calculateVertexNorm();
	
	//! makes lists with the vertexes and faces in the
	//  polygon model for displaying
	void makeDispObjects();

private:

	//! the number of tissue types
	int m_numTissueTypes;

	//! the number of faces in the differen tissue types
	TissueType *m_ptissue;

	//! Maximum number of characters in a file name (including the null).
    enum {FILENAMESIZE = 1024};

    // Maxi. numbers of chars in a line read from a file (incl. the null).
    enum {INPUTLINESIZE = 1024};

	//! read a number line from a infile
    char *readnumberline(char* string, FILE* infile, char* infilename);
	//! find the next number, skipping spaces and comments
    char *findnextnumber(char* string);

	//! tetrahedron list
	Tetrahedron *m_ptetList;
	//! number of tetrahedrons
	int m_numTet;
	//! number of tetrahedron attributes
	int m_numTetAttrib;

	//! vertex list
	Vertex *m_pvertexList; 
	//! number of vertexes
	int m_numVertexes;

	//! The connection between the Vertexes and 
	//! the tetrahedrons
	Vertet *m_pvertexTet;

	//! the dimention of the mesh
	int m_meshdim;
	//! the first vertex index in the file
	int m_firstnum;
};
#endif
