/**	tetrahedron data structure
	made by Rolf Anders Syvertsen
	2007-04-13
	rolfans@stud.ntnu.no
*/

#ifndef TETRATEDRON_H
#define TETRAHEDRON_H

#include "vertex.h"
/** The neighboring tetrahedrons:
	m_pfaceNeighb1= vertex0 + vertex1 + vertex2
	m_pfaceNeighb2= vertex0 + vertex2 + vertex3
	m_pfaceNeighb3= vertex0 + vertex3 + vertex1
	m_pfaceNeighb4= vertex1 + vertex3 + vertex2

*/
class Tetrahedron
{
public:
	//! constructor
	Tetrahedron();
	//! destructor
	~Tetrahedron();
	//! add the corners
	void addCorners(int verti1,int verti2, int verti3, int verti4);
	//! set the tissue type
	void setType(int type);
	//! get the tissue type
	int getType();
	//! get the vertex indexes
	int *getVertex();
	//! set face normals
	void setFaceNorm(int ind,Vertex norm);
	//! get Face Normals at
	Vertex getFaceNormAt(int index);

	//! set neighbor tet at ind 0-3
	//! sets the faces that should be rendered
	void setNeighbor(int ind, Tetrahedron *tet);
	//! how many neighbors are sett
	int numNeighbors();
	//! is the face at ind going to be rendered
	bool renderFaceAt(int ind);

private:
	//! Vertex indexes for the four vertexes
	int m_vertex[4];
	//! Face normals
	Vertex m_faceNormal[4];
	//! tetrahedron type
	int m_type;

	//! neighbor tetrahedrons
	Tetrahedron *m_pfaceNeighb1,*m_pfaceNeighb2,*m_pfaceNeighb3,*m_pfaceNeighb4;
	//! is the faces going to be rendered
	bool m_renderface[4];
	//! number of tet neighbors
	int m_neighbors;

};
#endif
